/* xdwapi sample 12

Function :
	Verifies signatures in the DocuWorks document, and acquires those information.
How to use :
	xdwsmpl12 file
*/

#include <stdio.h>
#include <windows.h>
#include <xdw_api.h>
#include <time.h> //for ctime()
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>

void check_signature(XDW_DOCUMENT_HANDLE handle, int sig);
void check_signature_stamp(XDW_DOCUMENT_HANDLE handle, int sig);
void check_signature_pki(XDW_DOCUMENT_HANDLE handle, int sig);

void print_sig_info(XDW_SIGNATURE_INFO_V5* sig_info);
void print_sig_stamp_info(XDW_SIGNATURE_STAMP_INFO_V5* stamp_info);
void print_sig_pki_info(XDW_SIGNATURE_PKI_INFO_V5* stamp_info);
void get_cert_data(XDW_DOCUMENT_HANDLE handle, int sig, int cert_size);
void print_error(int code);
void print_module_error(int module, int code);

// Print usage.
void print_usage()
{
	fprintf(stderr, "Usage : xdwsmpl12 file\n");
}


int main(int argc, char** argv)
{
	if (argc != 2) {
		print_usage();
		return 0;
	}

	// Open the document handle.
	XDW_DOCUMENT_HANDLE doc_handle;
	XDW_OPEN_MODE_EX open_mode;
	open_mode.nSize = sizeof(XDW_OPEN_MODE_EX);
	open_mode.nOption = XDW_OPEN_READONLY;
	open_mode.nAuthMode = XDW_AUTH_NODIALOGUE;
	int api_result = XDW_OpenDocumentHandle(argv[1], &doc_handle, (XDW_OPEN_MODE *)&open_mode );

	if ( api_result == 0 ) {
		// Acquires the number of signatures attached to a DocuWorks file.
		int sig_num = XDW_GetDocumentSignatureNumber(doc_handle, NULL);

		if ( sig_num >= 0 ) {
			fprintf(stdout, "The number of signatures : %d\n", sig_num);
			for ( int i = 1; i <= sig_num; i++ ) {
				// Acquires signature information.
				check_signature(doc_handle, i);
			}
		} else {
			print_error(api_result);
		}

		// Close the document handle.
		XDW_CloseDocumentHandle(doc_handle, NULL);
	} else {
		print_error(api_result);
		return 0;
	}
	
	return 1;
}

// Acquires signature infomation.
void check_signature(XDW_DOCUMENT_HANDLE handle, int sig)
{
	fprintf(stdout, "%dth signature :\n", sig);

	// Updates signature status.
	XDW_SIGNATURE_MODULE_STATUS module_status;
	module_status.nSize = sizeof(XDW_SIGNATURE_MODULE_STATUS);
	int api_result = XDW_UpdateSignatureStatus(handle, sig, NULL, NULL, &module_status);

	if ( api_result != 0 ) {
		// Print XDWAPI error.
		print_error(api_result);
		if (api_result == XDW_E_SIGNATURE_MODULE) {
			// Print error unique to the signature module.
			print_module_error(module_status.nSignatureType, module_status.nErrorStatus);
		}
	}

	// Acquires signature infomations.
	XDW_SIGNATURE_INFO_V5 sig_info;
	sig_info.nSize = sizeof(XDW_SIGNATURE_INFO_V5);
	api_result = XDW_GetSignatureInformation(handle, sig, &sig_info, NULL, NULL, &module_status);

	if ( api_result == 0 ) {
		print_sig_info(&sig_info);
		if (sig_info.nSignatureType == XDW_SIGNATURE_STAMP) {
			// Acquires infomation of DocuWorks electronic signature.
			check_signature_stamp(handle, sig);
		} else if (sig_info.nSignatureType == XDW_SIGNATURE_PKI) {
			// Acquires infomations of electronic certificate.
			check_signature_pki(handle, sig);
		}
	} else {
		print_error(api_result);
		if (api_result == XDW_E_SIGNATURE_MODULE) {
			print_module_error(module_status.nSignatureType, module_status.nErrorStatus);
		}
	}
}

// Acquires infomation of DocuWorks electronic signature.
void check_signature_stamp(XDW_DOCUMENT_HANDLE handle, int sig)
{
	// Specify pModuleInfo argument to acquire signature infomations.
	XDW_SIGNATURE_INFO_V5 sig_info;
	sig_info.nSize = sizeof(XDW_SIGNATURE_INFO_V5);
	XDW_SIGNATURE_STAMP_INFO_V5 stamp_info;
	stamp_info.nSize = sizeof(XDW_SIGNATURE_STAMP_INFO_V5);
	XDW_SIGNATURE_MODULE_STATUS module_status;
	module_status.nSize = sizeof(XDW_SIGNATURE_MODULE_STATUS);
	int api_result = XDW_GetSignatureInformation(handle, sig, &sig_info, &stamp_info, NULL, &module_status);

	if ( api_result == 0 ) {
		print_sig_stamp_info(&stamp_info);
	} else {
		print_error(api_result);
		if (api_result == XDW_E_SIGNATURE_MODULE) {
			print_module_error(module_status.nSignatureType, module_status.nErrorStatus);
		}
	}
}

// Acquires infomations of electronic certificate.
void check_signature_pki(XDW_DOCUMENT_HANDLE handle, int sig)
{
	// Specify pModuleInfo argument to acquire signature infomations.
	XDW_SIGNATURE_INFO_V5 sig_info;
	sig_info.nSize = sizeof(XDW_SIGNATURE_INFO_V5);
	XDW_SIGNATURE_PKI_INFO_V5 pki_info;
	pki_info.nSize = sizeof(XDW_SIGNATURE_PKI_INFO_V5);
	pki_info.pSignerCert = NULL;
	XDW_SIGNATURE_MODULE_STATUS module_status;
	module_status.nSize = sizeof(XDW_SIGNATURE_MODULE_STATUS);
	int api_result = XDW_GetSignatureInformation(handle, sig, &sig_info, &pki_info, NULL, &module_status);

	if ( api_result == 0 ) {
		print_sig_pki_info(&pki_info);
		if ( pki_info.nSignerCertSize > 0 ) {
			get_cert_data(handle, sig, pki_info.nSignerCertSize);
		}
	} else {
		print_error(api_result);
		if (api_result == XDW_E_SIGNATURE_MODULE) {
			print_module_error(module_status.nSignatureType, module_status.nErrorStatus);
		}
	}
}

// The information of certificate used for the signature is written out as a security certificate file.
void get_cert_data(XDW_DOCUMENT_HANDLE handle, int sig, int cert_size)
{
	// Allocate buffer to write certificate data.
	unsigned char* cert_data = (unsigned char*)new unsigned char[cert_size];
	if ( cert_data != NULL ) {
		// Specify the address to write certificate data and the size of buffer, then acquire signature informations.
		XDW_SIGNATURE_INFO_V5 sig_info;
		sig_info.nSize = sizeof(XDW_SIGNATURE_INFO_V5);
		XDW_SIGNATURE_PKI_INFO_V5 pki_info;
		pki_info.nSize = sizeof(XDW_SIGNATURE_PKI_INFO_V5);
		pki_info.nSignerCertSize = cert_size;
		pki_info.pSignerCert = cert_data;
		XDW_SIGNATURE_MODULE_STATUS module_status;
		module_status.nSize = sizeof(XDW_SIGNATURE_MODULE_STATUS);
		int api_result = XDW_GetSignatureInformation(handle, sig, &sig_info, &pki_info, NULL, &module_status);

		if ( api_result == 0 ) {
			// Certificate data is saved as a .cer file.
			char fname[256];
			sprintf(fname, "cert-%04d.cer", sig);
			int fh = _open(fname, _O_WRONLY | _O_CREAT | _O_BINARY, _S_IWRITE);
			if ( fh >= 0 ) {
				_write(fh, cert_data, cert_size);
				_close(fh);
			}
		} else {
			print_error(api_result);
			if (api_result == XDW_E_SIGNATURE_MODULE) {
				print_module_error(module_status.nSignatureType, module_status.nErrorStatus);
			}
		}
	}
}

// Print informations of the signature module.
void print_sig_info(XDW_SIGNATURE_INFO_V5* sig_info)
{
	fprintf(stdout, "  Signature module : ");
	switch (sig_info->nSignatureType) {
	case XDW_SIGNATURE_STAMP:
		fprintf(stdout, "DocuWorks electronic signature\n");
		break;
	case XDW_SIGNATURE_PKI:
		fprintf(stdout, "Electronic Certificate\n");
		break;
	default:
		fprintf(stdout, "Unknown module\n");
		break;
	}
	fprintf(stdout, "   Position : %d page(s). %.2fmm from the left. %.2fmm from the top.\n",
			sig_info->nPage, sig_info->nHorPos/100.0, sig_info->nVerPos/100.0);
	fprintf(stdout, "   Size : Width %.2fmm Height %.2fmm\n",
			sig_info->nWidth/100.0, sig_info->nHeight/100.0);
	fprintf(stdout, "   Date and Time: %s", ctime(&sig_info->nSignedTime));
}

// Print informations unique to the DocuWorks electronic signature.
void print_sig_stamp_info(XDW_SIGNATURE_STAMP_INFO_V5* stamp_info)
{
	fprintf(stdout, "    Signature Name : %s\n", stamp_info->lpszStampName);
	fprintf(stdout, "    Owner Name : %s\n", stamp_info->lpszOwnerName);
	fprintf(stdout, "    Expiration Date: %s", ctime(&stamp_info->nValidDate));
	fprintf(stdout, "    Remarks : %s\n", stamp_info->lpszRemarks);
	fprintf(stdout, "    The verification result of the document : ");
	switch (stamp_info->nDocVerificationStatus) {
	case XDW_SIGNATURE_STAMP_DOC_NONE:
		fprintf(stdout, "Validity Unknown\n");
		break;
	case XDW_SIGNATURE_STAMP_DOC_NOEDIT:
		fprintf(stdout, "Not modified\n");
		break;
	case XDW_SIGNATURE_STAMP_DOC_EDIT:
		fprintf(stdout, "Modified\n");
		break;
	case XDW_SIGNATURE_STAMP_DOC_BAD:
		fprintf(stdout, "Problem\n");
		break;
	default:
		fprintf(stdout, "Unknown status\n");
		break;
	}
	fprintf(stdout, "    The verification result of the Electronic Signature : ");
	switch (stamp_info->nStampVerificationStatus) {
	case XDW_SIGNATURE_STAMP_DOC_NONE:
		fprintf(stdout, "Validity Unknown\n");
		break;
	case XDW_SIGNATURE_STAMP_STAMP_TRUSTED:
		fprintf(stdout, "Trustworthy\n");
		break;
	case XDW_SIGNATURE_STAMP_STAMP_NOTRUST:
		fprintf(stdout, "No trustworthy\n");
		break;
	default:
		fprintf(stdout, "Unknown status\n");
		break;
	}
}

// Print informations unique to the electronic certificate.
void print_sig_pki_info(XDW_SIGNATURE_PKI_INFO_V5* pki_info)
{
	fprintf(stdout, "    Module Name : %s\n", pki_info->lpszModule);
	fprintf(stdout, "    Signer : %s (%s)\n", pki_info->lpszSubject, pki_info->lpszSubjectDN);
	fprintf(stdout, "    Certification Authority : %s (%s)\n", pki_info->lpszIssuer, pki_info->lpszIssuerDN);
	fprintf(stdout, "    Expiration Date : %s - %s\n", pki_info->lpszNotBefore, pki_info->lpszNotAfter);
	fprintf(stdout, "    Serial Number : %s\n", pki_info->lpszSerial);
	fprintf(stdout, "    Remarks : %s\n", pki_info->lpszRemarks);
	fprintf(stdout, "    Signature time : %s\n", pki_info->lpszSigningTime);
	fprintf(stdout, "    The verification result of the document : ");
	switch (pki_info->nDocVerificationStatus) {
	case XDW_SIGNATURE_PKI_DOC_UNKNOWN:
		fprintf(stdout, "Not verified.\n");
		break;
	case XDW_SIGNATURE_PKI_DOC_GOOD:
		fprintf(stdout, "The DocuWorks file has not been edited. The trustworthiness of the Certificate used for the signature cannot be confirmed.\n");
		break;
	case XDW_SIGNATURE_PKI_DOC_MODIFIED:
		fprintf(stdout, "The DocuWorks file has been edited. The trustworthiness of the Certificate used for the signature cannot be confirmed.\n");
		break;
	case XDW_SIGNATURE_PKI_DOC_BAD:
		fprintf(stdout, "The DocuWorks file has one or more problems. Or the Certificate used for the signature is invalid.\n");
		break;
	case XDW_SIGNATURE_PKI_DOC_GOOD_TRUSTED:
		fprintf(stdout, "The DocuWorks file has not been edited. The trustworthiness of the Certificate used for the signature has been confirmed.\n");
		break;
	case XDW_SIGNATURE_PKI_DOC_MODIFIED_TRUSTED:
		fprintf(stdout, "The DocuWorks file has been edited. The trustworthiness of the Certificate used for the signature has been confirmed.\n");
		break;
	default:
		fprintf(stdout, "Unknown status.\n");
		break;
	}
	fprintf(stdout, "    Verification Level : ");
	switch (pki_info->nCertVerificationType) {
	case XDW_SIGNATURE_PKI_TYPE_LOW:
		fprintf(stdout, "The trustworthiness level of the Certificate is low.\n");
		break;
	case XDW_SIGNATURE_PKI_TYPE_MID_LOCAL:
		fprintf(stdout, "The trustworthiness level of the Certificate is middle. Revocation is checked offline.\n");
		break;
	case XDW_SIGNATURE_PKI_TYPE_MID_NETWORK:
		fprintf(stdout, "The trustworthiness level of the Certificate is middle. Revocation is checked online.\n");
		break;
	case XDW_SIGNATURE_PKI_TYPE_HIGH_LOCAL:
		fprintf(stdout, "The trustworthiness level of the Certificate is high. Revocation is checked offline.\n");
		break;
	case XDW_SIGNATURE_PKI_TYPE_HIGH_NETWORK:
		fprintf(stdout, "The trustworthiness level of the Certificate is high. Revocation is checked online.\n");
		break;
	default:
		fprintf(stdout, "Unknown level.\n");
		break;
	}
	fprintf(stdout, "    The verification result of the certificate : ");
	switch (pki_info->nCertVerificationStatus) {
	case XDW_SIGNATURE_PKI_CERT_UNKNOWN:
		fprintf(stdout, "Not verified.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_OK:
		fprintf(stdout, "Valid.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_NO_ROOT_CERTIFICATE:
		fprintf(stdout, "Cannot find Root Certificate.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_NO_REVOCATION_CHECK:
		fprintf(stdout, "Failed to confirm revocation.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_OUT_OF_VALIDITY:
		fprintf(stdout, "Outside of the valid date.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_OUT_OF_VALIDITY_AT_SIGNED_TIME:
		fprintf(stdout, "Date of signature is outside of the valid date of the Certificate.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_REVOKE_CERTIFICATE:
		fprintf(stdout, "The Certificate is revoked.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_REVOKE_INTERMEDIATE_CERTIFICATE:
		fprintf(stdout, "The Intermediate Certificate is revoked.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_INVALID_SIGNATURE:
		fprintf(stdout, "The signature of the Certificate is invalid.\n");
		break;
	case XDW_SIGNATURE_PKI_CERT_INVALID_USAGE:
		fprintf(stdout, "The use of the key is invalid.\n");
		break;
	default:
		fprintf(stdout, "Unknown status.\n");
		break;
	}
}

// Print XDWAPI error.
void print_error(int code)
{
	switch (code) {
	case XDW_E_NOT_INSTALLED:
		fprintf(stderr, "DocuWorks not installed.\n");
		break;
	case XDW_E_FILE_NOT_FOUND:
		fprintf(stderr, "File not found.\n");
		break;
	case XDW_E_ACCESSDENIED:
	case XDW_E_INVALID_NAME:
	case XDW_E_BAD_NETPATH:
		fprintf(stderr, "File not opened.\n");
		break;
	case XDW_E_BAD_FORMAT:
		fprintf(stderr, "File bad format.\n");
		break;
	case XDW_E_INVALID_ACCESS:
		fprintf(stderr, "Invalid access.\n");
		break;
	case XDW_E_SIGNATURE_MODULE:
		fprintf(stderr, "Error unique to the signature module occurred.\n");
		break;
	default:
		fprintf(stderr, "Error.\n");
		break;
	}
}

// Print error unique to the signature module.
void print_module_error(int module, int code)
{
	switch (module) {
	// DocuWorks Electronic Signature
	case XDW_SIGNATURE_STAMP:
		fprintf(stderr, "DocuWorks Electronic Signature : ");
		switch (code) {
		case XDW_SIGNATURE_STAMP_ERROR_NO_OPENING_CASE:
			fprintf(stderr, "The processing failed because no open Electronic Signature Case exists.\n");
			break;
		case XDW_SIGNATURE_STAMP_ERROR_NO_SELFSTAMP:
			fprintf(stderr, "My Electronic Signature is not available in the Electronic Signature Case.\n");
			break;
		case XDW_SIGNATURE_STAMP_ERROR_OUT_OF_VALIDITY:
			fprintf(stderr, "The Electronic Signature is outside the valid date.\n");
			break;
		case XDW_SIGNATURE_STAMP_ERROR_INVALID_DATA:
			fprintf(stderr, "The processing failed because of uncomformity of data such as document to attach the signature and Electronic Signature Case.\n");
			break;
		case XDW_SIGNATURE_STAMP_ERROR_OUT_OF_MEMORY:
			fprintf(stderr, "The processing failed due to insufficient memory.\n");
			break;
		case XDW_SIGNATURE_STAMP_ERROR_UNKNOWN:
			fprintf(stderr, "The processing failed due to unexpected error.\n");
			break;
		case XDW_SIGNATURE_STAMP_ERROR_OK:
			fprintf(stderr, "The processing succeeded.\n");
			break;
		default:
			fprintf(stderr, "Error.\n");
			break;
		}
		break;
	// Electronic Certificate
	case XDW_SIGNATURE_PKI:
		fprintf(stderr, "Electronic Certificate module : ");
		switch (code) {
		case XDW_SIGNATURE_PKI_ERROR_BAD_PLATFORM:
			fprintf(stderr, "Executing environment is inappropriate.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_WRITE_REG_ERROR:
			fprintf(stderr, "Writing operation to the registry failed.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_TRUST_LEVEL:
			fprintf(stderr, "The set trustworthiness level is invalid.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_REVOKE_CHECK_TYPE:
			fprintf(stderr, "The set verification method is invalid.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_AUTO_IMPORT_CERT_FLAG:
			fprintf(stderr, "The value of auto registration flag of the Certificate is invalid.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_SIGN_CONFIG:
			fprintf(stderr, "The setting of the signature is invalid.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_NO_IMAGE_FILE:
			fprintf(stderr, "Image file does not exist.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_SIGN_CERT:
			fprintf(stderr, "The selected Certificate can no longer be used.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_NO_SIGN_CERT:
			fprintf(stderr, "The Certificate for signature is not selected.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_NOT_USE_PRIVATE_KEY:
			fprintf(stderr, "Access to Private key failed.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_INVALID:
			fprintf(stderr, "The Certificate is already revoked.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_SIGN:
			fprintf(stderr, "The signature of the Certificate is invalid.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_REVOKE_CHECK_ERROR:
			fprintf(stderr, "Revocation of the Certificate cannot be confirmed.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_OUT_OF_VALIDITY:
			fprintf(stderr, "The Certificate is outside the valid date.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_NO_CERT:
			fprintf(stderr, "The Certificate does not exist.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_FAILURE_IMPOPT_CERT:
			fprintf(stderr, "Registration of the Certificate failed.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_NO_ROOT_CERT:
			fprintf(stderr, "The Root Certificate does not exist.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_CERT_SIZE:
			fprintf(stderr, "The Certificate size is invalid.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_BAD_ARG:
			fprintf(stderr, "The argument is invalid.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_UNKNOWN:
			fprintf(stderr, "Unknown error.\n");
			break;
		case XDW_SIGNATURE_PKI_ERROR_OK:
			fprintf(stderr, "The processing succeeded.\n");
			break;
		default:
			fprintf(stderr, "Error.\n");
			break;
		}
		break;
	default:
		fprintf(stderr, "Unknown signature module.\n");
		break;
	}
}
